RESERVED_ENV_VAR_NAMES = [
    "DAGSTER_CLOUD_IS_BRANCH_DEPLOYMENT",
    "DAGSTER_CLOUD_DEPLOYMENT_NAME",
    "DAGSTER_CLOUD_LOCATION_NAME",
    "DAGSTER_CLOUD_GIT_SHA",
    "DAGSTER_CLOUD_GIT_TIMESTAMP",
    "DAGSTER_CLOUD_GIT_AUTHOR_EMAIL",
    "DAGSTER_CLOUD_GIT_AUTHOR_NAME",
    "DAGSTER_CLOUD_GIT_MESSAGE",
    "DAGSTER_CLOUD_GIT_BRANCH",
    "DAGSTER_CLOUD_GIT_REPO",
    "DAGSTER_CLOUD_PULL_REQUEST_ID",
    "DAGSTER_CLOUD_PULL_REQUEST_STATUS",
]



import os
from collections.abc import Sequence
from typing import IO, TYPE_CHECKING, Any, Optional

import requests
from dagster import (
    Field,
    IntSource,
    StringSource,
    _check as check,
)
from dagster._core.storage.cloud_storage_compute_log_manager import (
    TruncatingCloudStorageComputeLogManager,
)
from dagster._core.storage.compute_log_manager import ComputeIOType
from dagster._core.storage.local_compute_log_manager import (
    IO_TYPE_EXTENSION,
    LocalComputeLogManager,
)
from dagster._serdes import ConfigurableClass, ConfigurableClassData
from dagster_cloud_cli.core.errors import raise_http_error
from dagster_cloud_cli.core.headers.auth import DagsterCloudInstanceScope
from dagster_shared import seven
from requests.adapters import HTTPAdapter
from typing_extensions import Self

if TYPE_CHECKING:
    from dagster_cloud.instance import DagsterCloudAgentInstance  # noqa: F401


class CloudComputeLogManager(
    TruncatingCloudStorageComputeLogManager["DagsterCloudAgentInstance"], ConfigurableClass
):
    def __init__(
        self,
        local_dir=None,
        upload_interval=None,
        inst_data=None,
    ):
        # proxy calls to local compute log manager (for subscriptions, etc)
        if not local_dir:
            local_dir = seven.get_system_temp_directory()

        self._upload_session = requests.Session()
        adapter = HTTPAdapter(max_retries=3)
        self._upload_session.mount("http://", adapter)
        self._upload_session.mount("https://", adapter)

        self._local_manager = LocalComputeLogManager(local_dir)
        self._upload_interval = check.opt_int_param(upload_interval, "upload_interval")
        self._inst_data = check.opt_inst_param(inst_data, "inst_data", ConfigurableClassData)
        super().__init__()

    @property
    def inst_data(self):
        return self._inst_data

    @classmethod
    def config_type(cls):
        return {
            "local_dir": Field(StringSource, is_required=False),
            "upload_interval": Field(IntSource, is_required=False),
        }

    @classmethod
    def from_config_value(cls, inst_data: ConfigurableClassData, config_value: Any) -> Self:
        return cls(inst_data=inst_data, **config_value)

    @property
    def local_manager(self) -> LocalComputeLogManager:
        return self._local_manager

    @property
    def upload_interval(self) -> Optional[int]:
        return self._upload_interval

    def delete_logs(
        self, log_key: Optional[Sequence[str]] = None, prefix: Optional[Sequence[str]] = None
    ):
        raise NotImplementedError("User Agent should not need to delete compute logs")

    def download_url_for_type(self, log_key: Sequence[str], io_type: ComputeIOType):
        raise NotImplementedError("User Agent should not need to download compute logs")

    def display_path_for_type(self, log_key: Sequence[str], io_type: ComputeIOType):
        raise NotImplementedError("User Agent should not need to download compute logs")

    def cloud_storage_has_logs(
        self, log_key: Sequence[str], io_type: ComputeIOType, partial: bool = False
    ) -> bool:
        """Returns whether the cloud storage contains logs for a given log key."""
        return False

    def _upload_file_obj(
        self, data: IO[bytes], log_key: Sequence[str], io_type: ComputeIOType, partial=False
    ):
        path = self.local_manager.get_captured_local_path(log_key, IO_TYPE_EXTENSION[io_type])
        size_bytes = os.stat(path).st_size
        params: dict[str, Any] = {
            "log_key": log_key,
            "io_type": io_type.value,
            "size_bytes": size_bytes,
            # for back-compat
            "run_id": log_key[0],
            "key": log_key[-1],
            "method": "PUT",
        }
        if partial:
            params["partial"] = True
        resp = self._instance.requests_managed_retries_session.get(
            self._instance.dagster_cloud_gen_logs_url_url,
            params=params,
            headers=self._instance.dagster_cloud_api_headers(DagsterCloudInstanceScope.DEPLOYMENT),
            timeout=self._instance.dagster_cloud_api_timeout,
            proxies=self._instance.dagster_cloud_api_proxies,
        )
        raise_http_error(resp)
        resp_data = resp.json()

        if resp_data.get("skip_upload"):
            return

        self._upload_session.put(
            resp_data["url"],
            data=data,
            timeout=self._instance.dagster_cloud_api_timeout,
        )

    def download_from_cloud_storage(
        self, log_key: Sequence[str], io_type: ComputeIOType, partial=False
    ):
        raise NotImplementedError("User Agent should not need to download compute logs")

    def on_subscribe(self, subscription):
        raise NotImplementedError("User Agent should not need to download compute logs")



import hashlib
import json
import os
import uuid
from typing import Optional

import yaml
from dagster._core.instance.ref import InstanceRef
from dagster._serdes import serialize_value
from dagster_cloud_cli.core.workspace import CodeLocationDeployData

DEFAULT_CODE_SERVER_PORT = 4000


def get_code_server_port() -> int:
    """Returns the code server port from DAGSTER_CLOUD_CODE_SERVER_PORT env var.

    Defaults to 4000 if the environment variable is not set.
    """
    port_str = os.environ.get("DAGSTER_CLOUD_CODE_SERVER_PORT")
    if port_str is None:
        return DEFAULT_CODE_SERVER_PORT
    return int(port_str)


def unique_resource_name(deployment_name, location_name, length_limit, sanitize_fn):
    hash_value = str(uuid.uuid4().hex)[0:6]

    name = f"{location_name}-{deployment_name}"
    sanitized_location_name = sanitize_fn(name) if sanitize_fn else name
    truncated_location_name = sanitized_location_name[: (length_limit - 7)]
    sanitized_unique_name = (
        f"{truncated_location_name}-{hash_value}" if truncated_location_name else hash_value
    )
    assert len(sanitized_unique_name) <= length_limit
    return sanitized_unique_name


def get_human_readable_label(name, length_limit, sanitize_fn):
    truncated_name = name[:length_limit]
    return sanitize_fn(truncated_name) if sanitize_fn else truncated_name


def deterministic_label_for_location(deployment_name: str, location_name: str) -> str:
    """Need a label here that is a unique function of location name since we use it to
    search for existing deployments on update and remove them. Does not need to be human-readable.
    """
    m = hashlib.sha1()  # Creates a 40-byte hash
    m.update(f"{deployment_name}-{location_name}".encode())

    unique_label = m.hexdigest()
    return unique_label


def get_instance_ref_for_user_code(instance_ref: InstanceRef) -> InstanceRef:
    # Remove fields from InstanceRef that may not be compatible with earlier
    # versions of dagster and aren't actually needed by user code

    custom_instance_class_data = instance_ref.custom_instance_class_data
    if custom_instance_class_data:
        config_dict = custom_instance_class_data.config_dict
        new_config_dict = {
            key: val
            for key, val in config_dict.items()
            if key
            not in {
                "agent_queues",
                "allowed_full_deployment_locations",
                "allowed_branch_deployment_locations",
            }
        }

        user_code_launcher_config = config_dict.get("user_code_launcher", {}).get("config")
        if user_code_launcher_config:
            new_config_dict["user_code_launcher"]["config"] = {
                key: val
                for key, val in user_code_launcher_config.items()
                if key not in {"agent_metrics"}
            }

        custom_instance_class_data = custom_instance_class_data._replace(
            config_yaml=yaml.dump(new_config_dict)
        )

    return instance_ref._replace(custom_instance_class_data=custom_instance_class_data)


def get_grpc_server_env(
    code_location_deploy_data: CodeLocationDeployData,
    port: Optional[int],
    location_name: str,
    instance_ref: Optional[InstanceRef],
    socket: Optional[str] = None,
) -> dict[str, str]:
    return {
        **{
            "DAGSTER_LOCATION_NAME": location_name,
            "DAGSTER_INJECT_ENV_VARS_FROM_INSTANCE": "1",
            "DAGSTER_CLI_API_GRPC_LAZY_LOAD_USER_CODE": "1",
            "DAGSTER_CLI_API_GRPC_HOST": "0.0.0.0",
            # include the container context on the code server to ensure that it is uploaded
            # to the server and included in the workspace snapshot, which ensures that while a
            # code location is still being updated, existing code will use the snapshotted
            # version rather than the newly updated version
            "DAGSTER_CONTAINER_CONTEXT": json.dumps(code_location_deploy_data.container_context),
        },
        **(
            {"DAGSTER_INSTANCE_REF": serialize_value(get_instance_ref_for_user_code(instance_ref))}
            if instance_ref
            else {}
        ),
        **({"DAGSTER_CLI_API_GRPC_PORT": str(port)} if port else {}),
        **({"DAGSTER_CLI_API_GRPC_SOCKET": str(socket)} if socket else {}),
        **(
            {"DAGSTER_CURRENT_IMAGE": code_location_deploy_data.image}
            if code_location_deploy_data.image
            else {}
        ),
        **(
            {"DAGSTER_CLI_API_GRPC_PYTHON_FILE": code_location_deploy_data.python_file}
            if code_location_deploy_data.python_file
            else {}
        ),
        **(
            {"DAGSTER_CLI_API_GRPC_MODULE_NAME": code_location_deploy_data.module_name}
            if code_location_deploy_data.module_name
            else {}
        ),
        **(
            {"DAGSTER_CLI_API_GRPC_PACKAGE_NAME": code_location_deploy_data.package_name}
            if code_location_deploy_data.package_name
            else {}
        ),
        **(
            {"DAGSTER_CLI_API_GRPC_WORKING_DIRECTORY": code_location_deploy_data.working_directory}
            if code_location_deploy_data.working_directory
            else {}
        ),
        **(
            {"DAGSTER_CLI_API_GRPC_ATTRIBUTE": code_location_deploy_data.attribute}
            if code_location_deploy_data.attribute
            else {}
        ),
        **(
            {"DAGSTER_CLI_API_GRPC_USE_PYTHON_ENVIRONMENT_ENTRY_POINT": "1"}
            if code_location_deploy_data.executable_path
            else {}
        ),
        **(
            {
                "DAGSTER_CLI_API_GRPC_AUTOLOAD_DEFS_MODULE_NAME": code_location_deploy_data.autoload_defs_module_name
            }
            if code_location_deploy_data.autoload_defs_module_name
            else {}
        ),
    }



import logging
from typing import Optional

from dagster._utils.cached_method import cached_method


class Service:
    def __init__(self, client, arn):
        self.client = client
        self.arn = self._long_arn(arn)
        self.name = arn.split("/")[-1]

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.arn == other.arn

    def __hash__(self) -> int:
        return hash(self.arn)

    def __str__(self) -> str:
        return self.arn

    @property
    def hostname(self):
        hostname = f"{self.name}.{self.client.namespace}"
        return hostname

    @property
    def tags(self):
        if not self.service_discovery_arn:
            return {}

        try:
            tags = self.client.service_discovery.list_tags_for_resource(
                ResourceARN=self.service_discovery_arn
            ).get("Tags")
        except self.client.service_discovery.exceptions.ResourceNotFoundException:
            logging.warning(f"Could not find service discovery ARN {self.service_discovery_arn}")
            return {}

        return dict([(tag["Key"], tag["Value"]) for tag in tags])

    @property
    @cached_method
    def service_discovery_arn(self):
        service = self.client.ecs.describe_services(
            cluster=self.client.cluster_name,
            services=[self.arn],
        ).get("services", [{}])[0]

        registries = service.get("serviceRegistries") or [{}]
        arn = registries[0].get("registryArn")

        return arn

    @property
    @cached_method
    def public_ip(self):
        task_arns = self.client.ecs.list_tasks(
            cluster=self.client.cluster_name,
            serviceName=self.name,
        ).get("taskArns")

        # Assume there's only one task per service
        task = self.client.ecs.describe_tasks(
            cluster=self.client.cluster_name,
            tasks=task_arns,
        ).get("tasks")[0]

        # Assume there's only one attachment per task
        attachment_details = task.get("attachments")[0].get("details")
        for detail in attachment_details:
            if detail.get("name") == "networkInterfaceId":
                eni_id = detail.get("value")
                eni = self.client.ec2.NetworkInterface(eni_id)
                if eni.association_attribute:
                    return eni.association_attribute.get("PublicIp")

    @property
    @cached_method
    def create_timestamp(self) -> Optional[float]:
        response = self.client.ecs.describe_services(
            cluster=self.client.cluster_name, services=[self.name]
        )

        service_description = response["services"][0]

        # Extract the creation timestamp
        return service_description["createdAt"].timestamp()

    @property
    @cached_method
    def environ(self):
        task_arns = self.client.ecs.list_tasks(
            cluster=self.client.cluster_name,
            serviceName=self.name,
        ).get("taskArns")

        # Assume there's only one task per service
        task = self.client.ecs.describe_tasks(
            cluster=self.client.cluster_name,
            tasks=task_arns,
        ).get("tasks")[0]

        # Assume there's only one container per task
        task_definition_arn = task.get("taskDefinitionArn")
        container = self.client.ecs.describe_task_definition(taskDefinition=task_definition_arn)[
            "taskDefinition"
        ]["containerDefinitions"][0]
        return dict((env["name"], env["value"]) for env in container["environment"])

    def _long_arn(self, arn):
        # https://docs.aws.amazon.com/AmazonECS/latest/userguide/ecs-account-settings.html#ecs-resource-ids
        arn_parts = arn.split("/")
        if len(arn_parts) == 3:
            # A new long arn:
            # arn:aws:ecs:region:aws_account_id:service/cluster-name/service-name
            # Return it as is
            return arn
        else:
            # An old short arn:
            # arn:aws:ecs:region:aws_account_id:service/service-name
            # Add the cluster name
            return "/".join([arn_parts[0], self.client.cluster_name, arn_parts[1]])



from typing import Optional, Union
from uuid import uuid4 as uuid

from dagster import (
    AssetExecutionContext,
    AssetKey,
    AssetObservation,
    JobDefinition,
    OpExecutionContext,
)

# Metadata key prefix used to tag Snowflake queries with opaque IDs
OPAQUE_ID_METADATA_KEY_PREFIX = "dagster_snowflake_opaque_id:"
OPAQUE_ID_SQL_SIGIL = "snowflake_dagster_dbt_v1_opaque_id"

OUTPUT_NON_ASSET_SIGIL = "__snowflake_query_metadata_"


def build_opaque_id_metadata(opaque_id: str) -> dict:
    return {f"{OPAQUE_ID_METADATA_KEY_PREFIX}{opaque_id}": True}


def meter_snowflake_query(
    context: Union[OpExecutionContext, AssetExecutionContext],
    sql: str,
    comment_factory=lambda comment: f"\n-- {comment}\n",
    opaque_id=None,
    associated_asset_key: Optional[AssetKey] = None,
):
    """A utility function that takes a SQL query and returns a modified version of the query
    that includes a comment that will be used to identify the query when attributing cost.
    Logs an AssetObservation event to associate the query with the executing op/asset.

    .. code-block:: python

        from dagster import asset, AssetExecutionContext
        from dagster_snowflake import SnowflakeResource
        from dagster_cloud.dagster_insights import meter_snowflake_query

        @asset
        def my_cool_asset(
            context: AssetExecutionContext, snowflake: SnowflakeResource
        ) -> Dict[str, Any]:
            with snowflake.get_connection() as conn:
                res = conn.execute_query(meter_snowflake_query(context, "SELECT * FROM my_big_table"), fetch_results=True)

            return res

    """
    if context.has_assets_def and len(context.assets_def.keys_by_output_name.keys()) == 1:
        inferred_asset_key = context.asset_key
        if associated_asset_key and associated_asset_key != inferred_asset_key:
            raise RuntimeError(
                f"Op materializes asset key {inferred_asset_key}, which does not match supplied"
                f" asset key {associated_asset_key}"
            )
        associated_asset_key = inferred_asset_key
    elif context.has_assets_def:
        asset_keys = context.assets_def.keys_by_output_name.values()

        if not associated_asset_key:
            context.log.warn(
                "Op materializes assets, but no asset key associated with Snowflake query usage."
            )
        if associated_asset_key and associated_asset_key not in asset_keys:
            raise RuntimeError(
                f"Op materializes assets {asset_keys}, but supplied asset key"
                f" {associated_asset_key} not found"
            )
    elif not context.has_assets_def and associated_asset_key:
        raise RuntimeError(
            f"Op does not materialize assets, but supplied asset key {associated_asset_key} found"
        )

    associated_output_name = None
    if associated_asset_key:
        associated_output_name = context.assets_def.get_output_name_for_asset_key(
            associated_asset_key
        )

    if opaque_id is None:
        opaque_id = str(uuid())
    modified_sql = sql + comment_factory(f"{OPAQUE_ID_SQL_SIGIL}[[[{opaque_id}]]]")

    if associated_output_name and associated_asset_key:
        context.log_event(
            AssetObservation(
                asset_key=associated_asset_key,
                metadata=build_opaque_id_metadata(opaque_id),
            )
        )
    else:
        context.log_event(
            AssetObservation(
                asset_key=marker_asset_key_for_job(context.job_def),
                metadata=build_opaque_id_metadata(opaque_id),
            )
        )

    return modified_sql


def marker_asset_key_for_job(
    job: JobDefinition,
) -> AssetKey:
    return AssetKey(path=[f"{OUTPUT_NON_ASSET_SIGIL}{job.name}"])



import logging
from collections import defaultdict
from collections.abc import Collection, Mapping
from contextlib import contextmanager
from pathlib import Path
from typing import Any, NamedTuple, Optional

import kubernetes
import kubernetes.client as client
from dagster import (
    Array,
    Field,
    IntSource,
    Map,
    Noneable,
    Permissive,
    Shape,
    StringSource,
    _check as check,
)
from dagster._serdes import ConfigurableClass
from dagster._serdes.config_class import ConfigurableClassData
from dagster._utils.merger import merge_dicts
from dagster_cloud_cli.core.workspace import CodeLocationDeployData
from dagster_k8s.client import PatchedApiClient
from dagster_k8s.container_context import K8sContainerContext
from dagster_k8s.job import UserDefinedDagsterK8sConfig
from dagster_k8s.models import k8s_snake_case_dict
from kubernetes.client.rest import ApiException
from typing_extensions import Self

from dagster_cloud.api.dagster_cloud_api import UserCodeDeploymentType
from dagster_cloud.constants import RESERVED_ENV_VAR_NAMES
from dagster_cloud.execution.cloud_run_launcher.k8s import CloudK8sRunLauncher
from dagster_cloud.execution.monitoring import CloudContainerResourceLimits
from dagster_cloud.workspace.kubernetes.utils import (
    construct_code_location_deployment,
    construct_code_location_service,
    get_deployment_failure_debug_info,
    unique_k8s_resource_name,
    wait_for_deployment_complete,
)
from dagster_cloud.workspace.user_code_launcher import (
    DEFAULT_SERVER_PROCESS_STARTUP_TIMEOUT,
    SHARED_USER_CODE_LAUNCHER_CONFIG,
    DagsterCloudGrpcServer,
    DagsterCloudUserCodeLauncher,
    ServerEndpoint,
    UserCodeLauncherEntry,
)
from dagster_cloud.workspace.user_code_launcher.utils import (
    deterministic_label_for_location,
    get_code_server_port,
)

DEFAULT_DEPLOYMENT_STARTUP_TIMEOUT = 300
DEFAULT_IMAGE_PULL_GRACE_PERIOD = 30

from dagster_cloud.workspace.config_schema.kubernetes import SHARED_K8S_CONFIG


class K8sHandle(NamedTuple):
    namespace: str
    name: str
    labels: Mapping[str, str]
    creation_timestamp: Optional[float]

    def __str__(self):
        return f"{self.namespace}/{self.name}"


class K8sUserCodeLauncher(DagsterCloudUserCodeLauncher[K8sHandle], ConfigurableClass):
    def __init__(
        self,
        dagster_home,
        instance_config_map,
        inst_data=None,
        namespace=None,
        kubeconfig_file=None,
        pull_policy=None,
        env_config_maps=None,
        env_secrets=None,
        env_vars=None,
        service_account_name=None,
        volume_mounts=None,
        volumes=None,
        image_pull_secrets=None,
        deployment_startup_timeout=None,
        image_pull_grace_period=None,
        labels=None,
        resources=None,
        scheduler_name=None,
        server_k8s_config=None,
        run_k8s_config=None,
        k8s_apps_api_client=None,
        k8s_core_api_client=None,
        security_context=None,
        only_allow_user_defined_k8s_config_fields=None,
        only_allow_user_defined_env_vars=None,
        fail_pod_on_run_failure=None,
        **kwargs,
    ):
        self._inst_data = inst_data
        self._logger = logging.getLogger("K8sUserCodeLauncher")
        self._dagster_home = check.str_param(dagster_home, "dagster_home")
        self._instance_config_map = check.str_param(instance_config_map, "instance_config_map")
        self._namespace = namespace

        self._pull_policy = pull_policy
        self._env_config_maps = check.opt_list_param(
            env_config_maps, "env_config_maps", of_type=str
        )
        self._env_secrets = check.opt_list_param(env_secrets, "env_secrets", of_type=str)
        self._env_vars = check.opt_list_param(env_vars, "env_vars", of_type=str)
        self._service_account_name = check.str_param(service_account_name, "service_account_name")
        self._volume_mounts = [
            k8s_snake_case_dict(kubernetes.client.V1VolumeMount, mount)
            for mount in check.opt_list_param(volume_mounts, "volume_mounts")
        ]
        self._volumes = [
            k8s_snake_case_dict(kubernetes.client.V1Volume, volume)
            for volume in check.opt_list_param(volumes, "volumes")
        ]
        self._image_pull_secrets = check.opt_list_param(
            image_pull_secrets, "image_pull_secrets", of_type=dict
        )
        self._deployment_startup_timeout = check.opt_int_param(
            deployment_startup_timeout,
            "deployment_startup_timeout",
            DEFAULT_DEPLOYMENT_STARTUP_TIMEOUT,
        )
        self._image_pull_grace_period = check.opt_int_param(
            image_pull_grace_period,
            "image_pull_grace_period",
            DEFAULT_IMAGE_PULL_GRACE_PERIOD,
        )
        self._labels = check.opt_dict_param(labels, "labels", key_type=str, value_type=str)
        self._resources = check.opt_dict_param(resources, "resources", key_type=str)

        self._scheduler_name = check.opt_str_param(scheduler_name, "scheduler_name")

        self._server_k8s_config = check.opt_dict_param(server_k8s_config, "server_k8s_config")
        self._run_k8s_config = check.opt_dict_param(run_k8s_config, "run_k8s_config")

        if kubeconfig_file:
            kubernetes.config.load_kube_config(kubeconfig_file)
        else:
            kubernetes.config.load_incluster_config()

        self._k8s_apps_api_client = k8s_apps_api_client
        self._k8s_core_api_client = k8s_core_api_client

        self._security_context = security_context

        if only_allow_user_defined_k8s_config_fields is not None:
            # Dagster Cloud always sets some environment variables whenever it launched code.
            # If only_allow_user_defined_k8s_config_fields is being used to lock down which
            # k8s config can be set by the control plane and "container_config.env" is not
            # in that allowlist, add it to that allowlist so that we don't fail every pod that
            # is launched, but restrict the set of env vars that can be configured to only
            # a fixed set of known Dagster Cloud environment variables, defined in code.

            if not only_allow_user_defined_k8s_config_fields.get("container_config", {}).get("env"):
                only_allow_user_defined_k8s_config_fields = {
                    **only_allow_user_defined_k8s_config_fields,
                    "container_config": {
                        **only_allow_user_defined_k8s_config_fields.get("container_config", {}),
                        "env": True,
                    },
                }

                if only_allow_user_defined_env_vars is None:
                    only_allow_user_defined_env_vars = []

        if only_allow_user_defined_env_vars is not None:
            only_allow_user_defined_env_vars = (
                only_allow_user_defined_env_vars + RESERVED_ENV_VAR_NAMES
            )

        self._only_allow_user_defined_k8s_config_fields = only_allow_user_defined_k8s_config_fields
        self._only_allow_user_defined_env_vars = only_allow_user_defined_env_vars

        super().__init__(**kwargs)

        self._launcher = CloudK8sRunLauncher(
            dagster_home=self._dagster_home,
            instance_config_map=self._instance_config_map,
            postgres_password_secret=None,
            job_image=None,
            image_pull_policy=self._pull_policy,
            image_pull_secrets=self._image_pull_secrets,
            service_account_name=self._service_account_name,
            env_config_maps=self._env_config_maps,
            env_secrets=self._env_secrets,
            env_vars=self._env_vars,
            job_namespace=self._namespace,  # pyright: ignore[reportArgumentType]
            volume_mounts=self._volume_mounts,
            volumes=self._volumes,
            labels=self._labels,
            resources=self._resources,
            scheduler_name=self._scheduler_name,
            run_k8s_config=self._run_k8s_config,
            fail_pod_on_run_failure=fail_pod_on_run_failure,
            kubeconfig_file=kubeconfig_file,
            load_incluster_config=not kubeconfig_file,
            security_context=self._security_context,
            # leave out the code server specific fields from only_allow_user_defined_k8s_config_fields
            only_allow_user_defined_k8s_config_fields=(
                {
                    key: val
                    for key, val in only_allow_user_defined_k8s_config_fields.items()
                    if key not in {"service_metadata", "deployment_metadata"}
                }
                if only_allow_user_defined_k8s_config_fields
                else only_allow_user_defined_k8s_config_fields
            ),
            only_allow_user_defined_env_vars=self._only_allow_user_defined_env_vars,
        )

        # mutable set of observed namespaces to assist with cleanup
        self._used_namespaces: dict[tuple[str, str], set[str]] = defaultdict(set)

    @property
    def requires_images(self):
        return True

    def register_instance(self, instance):
        super().register_instance(instance)
        self._launcher.register_instance(instance)

    @property
    def inst_data(self):
        return self._inst_data

    @classmethod
    def config_type(cls):
        container_context_config = SHARED_K8S_CONFIG.copy()
        del container_context_config["image_pull_policy"]  # uses 'pull_policy'
        del container_context_config["namespace"]  # default is different

        return merge_dicts(
            container_context_config,
            {
                "dagster_home": Field(StringSource, is_required=True),
                "instance_config_map": Field(StringSource, is_required=True),
                "kubeconfig_file": Field(StringSource, is_required=False),
                "deployment_startup_timeout": Field(
                    IntSource,
                    is_required=False,
                    default_value=DEFAULT_DEPLOYMENT_STARTUP_TIMEOUT,
                    description=(
                        "Timeout when creating a new Kubernetes deployment for a code server"
                    ),
                ),
                "server_process_startup_timeout": Field(
                    IntSource,
                    is_required=False,
                    default_value=DEFAULT_SERVER_PROCESS_STARTUP_TIMEOUT,
                    description=(
                        "Timeout when waiting for a code server to be ready after it is created"
                    ),
                ),
                "image_pull_grace_period": Field(
                    IntSource,
                    is_required=False,
                    default_value=DEFAULT_IMAGE_PULL_GRACE_PERIOD,
                ),
                "pull_policy": Field(
                    Noneable(StringSource),
                    is_required=False,
                    description="Image pull policy to set on launched Pods.",
                ),
                "namespace": Field(StringSource, is_required=False, default_value="default"),
                "scheduler_name": Field(StringSource, is_required=False),
                "run_k8s_config": Field(
                    Shape(
                        {
                            "container_config": Permissive(),
                            "pod_template_spec_metadata": Permissive(),
                            "pod_spec_config": Permissive(),
                            "job_config": Permissive(),
                            "job_metadata": Permissive(),
                            "job_spec_config": Permissive(),
                        }
                    ),
                    is_required=False,
                    description="Raw Kubernetes configuration for launched runs.",
                ),
                "fail_pod_on_run_failure": Field(
                    bool,
                    is_required=False,
                    description=(
                        "Whether the launched Kubernetes Jobs and Pods should fail if the Dagster"
                        " run fails"
                    ),
                ),
                "server_k8s_config": Field(
                    Shape(
                        {
                            "container_config": Permissive(),
                            "pod_spec_config": Permissive(),
                            "pod_template_spec_metadata": Permissive(),
                            "deployment_metadata": Permissive(),
                            "service_metadata": Permissive(),
                        }
                    ),
                    is_required=False,
                    description="Raw Kubernetes configuration for launched code servers.",
                ),
                "only_allow_user_defined_k8s_config_fields": Field(
                    Shape(
                        {
                            "container_config": Field(
                                Map(key_type=str, inner_type=bool), is_required=False
                            ),
                            "pod_spec_config": Field(
                                Map(key_type=str, inner_type=bool), is_required=False
                            ),
                            "pod_template_spec_metadata": Field(
                                Map(key_type=str, inner_type=bool), is_required=False
                            ),
                            "job_metadata": Field(
                                Map(key_type=str, inner_type=bool), is_required=False
                            ),
                            "job_spec_config": Field(
                                Map(key_type=str, inner_type=bool), is_required=False
                            ),
                            "deployment_metadata": Field(
                                Map(key_type=str, inner_type=bool), is_required=False
                            ),
                            "service_metadata": Field(
                                Map(key_type=str, inner_type=bool), is_required=False
                            ),
                            "namespace": Field(bool, is_required=False),
                        }
                    ),
                    is_required=False,
                    description="Dictionary of fields that are allowed to be configured on a "
                    "per-run or per-code-location basis - e.g. using tags on the run. "
                    "Can be used to prevent user code or the Dagster Cloud control plane from "
                    "being able to set arbitrary kubernetes config on the resources launched "
                    "by the agent.",
                ),
                "only_allow_user_defined_env_vars": Field(
                    Array(str),
                    is_required=False,
                    description="List of environment variable names that are allowed to be set on "
                    "a per-run or per-code-location basis - e.g. using tags on the run. ",
                ),
                "code_server_metrics": Field(
                    {"enabled": Field(bool, is_required=False, default_value=False)},
                    is_required=False,
                ),
                "agent_metrics": Field(
                    {"enabled": Field(bool, is_required=False, default_value=False)},
                    is_required=False,
                ),
            },
            SHARED_USER_CODE_LAUNCHER_CONFIG,
        )

    @property
    def user_code_deployment_type(self) -> UserCodeDeploymentType:
        return UserCodeDeploymentType.K8S

    @classmethod
    def from_config_value(cls, inst_data: ConfigurableClassData, config_value: Any) -> Self:
        return cls(
            inst_data=inst_data,
            **config_value,
        )

    def _get_core_api_client(self):
        return (
            self._k8s_core_api_client
            if self._k8s_core_api_client
            else client.CoreV1Api(api_client=PatchedApiClient())
        )

    @contextmanager
    def _get_apps_api_instance(self):
        if self._k8s_apps_api_client:
            yield self._k8s_apps_api_client
            return

        with PatchedApiClient() as api_client:
            yield client.AppsV1Api(api_client)

    def _resolve_container_context(self, metadata: CodeLocationDeployData) -> K8sContainerContext:
        user_defined_container_context = K8sContainerContext.create_from_config(
            metadata.container_context
        ).validate_user_k8s_config_for_code_server(
            self._only_allow_user_defined_k8s_config_fields, self._only_allow_user_defined_env_vars
        )
        return K8sContainerContext(
            image_pull_policy=self._pull_policy,
            image_pull_secrets=self._image_pull_secrets,
            service_account_name=self._service_account_name,
            env_config_maps=self._env_config_maps,
            env_secrets=self._env_secrets,
            env_vars=self._env_vars
            + [f"{k}={v}" for k, v in (metadata.cloud_context_env or {}).items()],
            volume_mounts=self._volume_mounts,
            volumes=self._volumes,
            labels=self._labels,
            namespace=self._namespace,
            resources=self._resources,
            scheduler_name=self._scheduler_name,
            security_context=self._security_context,
            server_k8s_config=UserDefinedDagsterK8sConfig.from_dict(self._server_k8s_config),
            run_k8s_config=UserDefinedDagsterK8sConfig.from_dict(self._run_k8s_config),
        ).merge(user_defined_container_context)

    def get_code_server_resource_limits(
        self, deployment_name: str, location_name: str
    ) -> CloudContainerResourceLimits:
        metadata = self._actual_entries[(deployment_name, location_name)].code_location_deploy_data
        k8s_container_context = metadata.container_context.get("k8s", {})
        resources = (
            k8s_container_context.get("server_k8s_config", {})
            .get("container_config", {})
            .get("resources", {})
        )
        if not resources:
            resources = k8s_container_context.get("resources", {})

        self._logger.info(
            f"Getting resource limits for deployment {deployment_name} in location {location_name}: {resources}"
        )
        return {
            "k8s": {
                "cpu_limit": resources.get("limits", {}).get("cpu"),
                "cpu_request": resources.get("requests", {}).get("cpu"),
                "memory_limit": resources.get("limits", {}).get("memory"),
                "memory_request": resources.get("requests", {}).get("memory"),
            }
        }

    def _start_new_server_spinup(
        self,
        deployment_name: str,
        location_name: str,
        desired_entry: UserCodeLauncherEntry,
    ) -> DagsterCloudGrpcServer:
        metadata = desired_entry.code_location_deploy_data

        args = metadata.get_grpc_server_command(
            metrics_enabled=self._instance.user_code_launcher.code_server_metrics_enabled
        )

        resource_name = unique_k8s_resource_name(deployment_name, location_name)

        container_context = self._resolve_container_context(metadata)

        deployment_reponse = None

        try:
            with self._get_apps_api_instance() as api_instance:
                deployment_reponse = api_instance.create_namespaced_deployment(
                    namespace=container_context.namespace,
                    body=construct_code_location_deployment(
                        self._instance,
                        deployment_name,
                        location_name,
                        resource_name,
                        metadata,
                        container_context,
                        args=args,
                        server_timestamp=desired_entry.update_timestamp,
                    ),
                )
            self._logger.info(
                f"Created deployment {deployment_reponse.metadata.name} in namespace {container_context.namespace}"
            )
        except ApiException as e:
            self._logger.error(
                f"Exception when calling AppsV1Api->create_namespaced_deployment: {e}\n"
            )
            raise e

        namespace = check.not_none(container_context.namespace)

        self._used_namespaces[(deployment_name, location_name)].add(namespace)

        try:
            service_response = self._get_core_api_client().create_namespaced_service(
                namespace,
                construct_code_location_service(
                    deployment_name,
                    location_name,
                    resource_name,
                    container_context,
                    self._instance,
                    server_timestamp=desired_entry.update_timestamp,
                ),
            )
            self._logger.info(
                f"Created service {service_response.metadata.name} in namespace {namespace}"
            )
        except ApiException as e:
            self._logger.error(
                f"Exception when calling AppsV1Api->create_namespaced_service: {e}\n"
            )
            raise e

        # use namespace scoped host name
        host = f"{resource_name}.{namespace}"

        endpoint = ServerEndpoint(
            host=host,
            port=get_code_server_port(),
            socket=None,
        )

        return DagsterCloudGrpcServer(
            K8sHandle(
                namespace=namespace,
                name=resource_name,
                labels=deployment_reponse.metadata.labels,
                creation_timestamp=deployment_reponse.metadata.creation_timestamp.timestamp()
                if deployment_reponse.metadata.creation_timestamp
                else None,
            ),
            endpoint,
            metadata,
        )

    def _get_timeout_debug_info(
        self,
        server_handle,
    ):
        core_api = self._get_core_api_client()
        k8s_deployment_name = server_handle.name
        namespace = server_handle.namespace

        pod_list = core_api.list_namespaced_pod(
            namespace, label_selector=f"user-deployment={k8s_deployment_name}"
        ).items

        with self._get_apps_api_instance() as apps_api_client:
            return get_deployment_failure_debug_info(
                k8s_deployment_name,
                namespace,
                core_api,
                pod_list,
                self._logger,
                apps_api_client=apps_api_client,
            )

    def _write_readiness_sentinel(self) -> None:
        # Write to a sentinel file to indicate that we've finished our initial
        # reconciliation - this is used to indicate that we're ready to
        # serve requests
        Path("/tmp/finished_initial_reconciliation_sentinel.txt").touch(exist_ok=True)
        self._logger.info(
            "Wrote readiness sentinel: indicating that agent is ready to serve requests"
        )

    async def _wait_for_new_server_ready(  # pyright: ignore[reportIncompatibleMethodOverride], fix me!
        self,
        deployment_name: str,
        location_name: str,
        user_code_launcher_entry: UserCodeLauncherEntry,
        server_handle: K8sHandle,
        server_endpoint: ServerEndpoint,
    ) -> None:
        metadata = user_code_launcher_entry.code_location_deploy_data

        await wait_for_deployment_complete(
            server_handle.name,
            server_handle.namespace,
            self._logger,
            location_name,
            metadata,
            timeout=self._deployment_startup_timeout,
            image_pull_grace_period=self._image_pull_grace_period,
            core_api=self._get_core_api_client(),
        )

        await self._wait_for_dagster_server_process(
            client=server_endpoint.create_client(),
            timeout=self._server_process_startup_timeout,
            get_timeout_debug_info=lambda: self._get_timeout_debug_info(server_handle),
        )

    def _get_standalone_dagster_server_handles_for_location(
        self,
        deployment_name: str,
        location_name: str,
    ) -> Collection[K8sHandle]:
        handles = []
        namespaces_to_search = self._used_namespaces.get(
            (deployment_name, location_name),
            [self._namespace],
        )
        with self._get_apps_api_instance() as api_instance:
            for namespace in namespaces_to_search:
                deployments = api_instance.list_namespaced_deployment(
                    namespace,
                    label_selector=(
                        f"location_hash={deterministic_label_for_location(deployment_name, location_name)},agent_id={self._instance.instance_uuid}"
                    ),
                ).items
                for deployment in deployments:
                    handles.append(
                        K8sHandle(
                            namespace=namespace,  # pyright: ignore[reportArgumentType]
                            name=deployment.metadata.name,
                            labels=deployment.metadata.labels,
                            creation_timestamp=deployment.metadata.creation_timestamp.timestamp()
                            if deployment.metadata.creation_timestamp
                            else None,
                        )
                    )

        return handles

    def _list_server_handles(self) -> list[K8sHandle]:
        namespaces: set[str] = set()
        if self._namespace:
            namespaces.add(self._namespace)
        for namespace_items in self._used_namespaces.values():
            for namespace in namespace_items:
                namespaces.add(namespace)
        handles: list[K8sHandle] = []
        with self._get_apps_api_instance() as api_instance:
            for namespace in namespaces:
                deployments = api_instance.list_namespaced_deployment(
                    namespace,
                    label_selector="managed_by=K8sUserCodeLauncher",
                ).items
                for deployment in deployments:
                    handles.append(
                        K8sHandle(
                            namespace,
                            deployment.metadata.name,
                            deployment.metadata.labels,
                            deployment.metadata.creation_timestamp.timestamp()
                            if deployment.metadata.creation_timestamp
                            else None,
                        )
                    )
        self._logger.info(f"Listing server handles: {handles}")
        return handles

    def get_agent_id_for_server(self, handle: K8sHandle) -> Optional[str]:
        return handle.labels.get("agent_id")

    def get_server_create_timestamp(self, handle: K8sHandle) -> Optional[float]:
        return handle.creation_timestamp

    def _remove_server_handle(self, server_handle: K8sHandle) -> None:
        # Since we track which servers to delete by listing the k8s deployments,
        # delete the k8s service first to ensure that it can't be left dangling
        try:
            self._get_core_api_client().delete_namespaced_service(
                server_handle.name, server_handle.namespace
            )
        except ApiException as e:
            if e.status == 404:
                self._logger.exception(
                    f"Tried to delete service {server_handle.name} but it was not found"
                )
            else:
                raise

        with self._get_apps_api_instance() as api_instance:
            try:
                api_instance.delete_namespaced_deployment(
                    server_handle.name, server_handle.namespace
                )
            except ApiException as e:
                if e.status == 404:
                    self._logger.exception(
                        f"Tried to delete deployment {server_handle.name} but it was not found"
                    )
                else:
                    raise

        self._logger.info(
            f"Removed deployment and service {server_handle.name} in namespace {server_handle.namespace}"
        )

    def __exit__(self, exception_type, exception_value, traceback):
        super().__exit__(exception_value, exception_value, traceback)
        self._launcher.dispose()

    def run_launcher(self):
        return self._launcher



import os
from collections.abc import Sequence
from pathlib import Path
from typing import TYPE_CHECKING, Optional, Union

from dagster import DagsterInvariantViolationError
from dagster._annotations import beta
from dagster._core.definitions.metadata import (
    AnchorBasedFilePathMapping,
    link_code_references_to_git,
)
from dagster._core.definitions.metadata.source_code import FilePathMapping
from dagster._core.types.loadable_target_origin import LoadableTargetOrigin

if TYPE_CHECKING:
    from dagster import AssetsDefinition, AssetSpec, SourceAsset
    from dagster._core.definitions.assets.definition.cacheable_assets_definition import (
        CacheableAssetsDefinition,
    )

import sys


def _locate_git_root() -> Optional[Path]:
    try:
        code_origin = LoadableTargetOrigin.get()
    except DagsterInvariantViolationError:
        return None

    # get module matching code_origin.module_name
    module_or_pkg_name = code_origin.module_name or code_origin.package_name
    if module_or_pkg_name:
        module = sys.modules.get(module_or_pkg_name)
        if module:
            code_origin_filepath = module.__file__
    elif code_origin.python_file:
        code_origin_filepath = code_origin.python_file

    if not code_origin_filepath:  # pyright: ignore[reportPossiblyUnboundVariable]
        return None
    current_dir = Path(code_origin_filepath)
    for parent in current_dir.parents:
        if (parent / ".git").exists():
            return parent
    return None


@beta
def link_code_references_to_git_if_cloud(
    assets_defs: Sequence[
        Union["AssetsDefinition", "SourceAsset", "CacheableAssetsDefinition", "AssetSpec"]
    ],
    git_url: Optional[str] = None,
    git_branch: Optional[str] = None,
    file_path_mapping: Optional[FilePathMapping] = None,
) -> Sequence[Union["AssetsDefinition", "SourceAsset", "CacheableAssetsDefinition", "AssetSpec"]]:
    """Wrapper function which converts local file path code references to hosted git URLs
    if running in a Dagster Plus cloud environment. This is determined by the presence of
    the `DAGSTER_CLOUD_DEPLOYMENT_NAME` environment variable. When running in any other context,
    the local file references are left as is.

    Args:
        assets_defs (Sequence[Union[AssetsDefinition, SourceAsset, CacheableAssetsDefinition]]):
            The asset definitions to which source control metadata should be attached.
            Only assets with local file code references (such as those created by
            `with_source_code_references`) will be converted.
        git_url (Optional[str]): Override base URL for the source control system. By default,
            inferred from the `DAGSTER_CLOUD_GIT_URL` environment variable provided by cloud.
            For example, "https://github.com/dagster-io/dagster".
        git_branch (str): Override branch in the source control system, such as "master".
            Defaults to the `DAGSTER_CLOUD_GIT_SHA` or `DAGSTER_CLOUD_GIT_BRANCH` environment variable.
        file_path_mapping (Optional[FilePathMapping]):
            Specifies the mapping between local file paths and their corresponding paths in a source control repository.
            If None, the function will attempt to infer the git root of the repository and use that as a reference
            point. Simple usage is to provide a `AnchorBasedFilePathMapping` instance, which specifies an anchor file in the
            repository and the corresponding local file path, which is extrapolated to all other local file paths.
            Alternatively, a mapping can be provided which takes a local file path and returns the corresponding path in
            the repository, allowing for more complex mappings.


    Examples:
        Basic usage:
        .. code-block:: python
                defs = Definitions(
                    assets=link_code_references_to_git_if_cloud(
                        with_source_code_references([my_dbt_assets]),
                    )
                )

        Provide custom mapping for repository file paths, e.g. if local files no longer match
        the repository structure:
        .. code-block:: python
                defs = Definitions(
                    assets=link_code_references_to_git_if_cloud(
                        with_source_code_references([my_dbt_assets]),
                        file_path_mapping=AnchorBasedFilePathMapping(
                            local_file_anchor=Path(__file__),
                            file_anchor_path_in_repository="python_modules/my_module/my-module/__init__.py",
                        ),
                    )
                )
    """
    is_dagster_cloud = os.getenv("DAGSTER_CLOUD_DEPLOYMENT_NAME") is not None

    if not is_dagster_cloud:
        return assets_defs

    git_url = git_url or os.getenv("DAGSTER_CLOUD_GIT_URL")
    git_branch = (
        git_branch or os.getenv("DAGSTER_CLOUD_GIT_SHA") or os.getenv("DAGSTER_CLOUD_GIT_BRANCH")
    )

    if not git_url or not git_branch:
        raise ValueError(
            "Detected that this is a Dagster Cloud deployment, but"
            " could not infer source control information for this repository. Please provide"
            " values for `git_url` and `git_branch`."
        )

    if not file_path_mapping:
        git_root = _locate_git_root()
        if not git_root:
            raise ValueError(
                "Detected that this is a Dagster Cloud deployment, but"
                " could not infer the git root for the repository. Please provide a value for "
                "`file_path_mapping`."
            )
        file_path_mapping = AnchorBasedFilePathMapping(
            local_file_anchor=git_root,
            file_anchor_path_in_repository="",
        )

    return link_code_references_to_git(
        assets_defs=assets_defs,
        git_url=git_url,
        git_branch=git_branch,
        file_path_mapping=file_path_mapping,
    )



GET_LATEST_DEFS_STATE_INFO_QUERY = """
    query getLatestDefsStateInfo {
        latestDefsStateInfo
    }
"""

SET_LATEST_VERSION_MUTATION = """
    mutation setLatestDefsStateVersion($key: String!, $version: String!) {
        defsState {
            setLatestDefsStateVersion(key: $key, version: $version) {
                ok
            }
        }
    }
"""



import hashlib
import re
from typing import Optional

from dagster._core.remote_origin import RemoteJobOrigin
from dagster_aws.ecs.utils import sanitize_family

from dagster_cloud.workspace.user_code_launcher.utils import (
    get_human_readable_label,
    unique_resource_name,
)


def unique_ecs_resource_name(deployment_name, location_name):
    return unique_resource_name(
        deployment_name,
        location_name,
        length_limit=60,  # Service discovery name must satisfy DNS
        sanitize_fn=lambda name: re.sub("[^a-z0-9-]", "", name).strip("-"),
    )


def get_ecs_human_readable_label(name):
    return get_human_readable_label(
        name,
        length_limit=60,  # Service discovery name must satisfy DNS
        sanitize_fn=lambda name: re.sub("[^a-z0-9-]", "", name).strip("-"),
    )


def _get_family_hash(name, max_length=32, hash_size=8):
    m = hashlib.sha1()
    m.update(name.encode("utf-8"))
    name_hash = m.hexdigest()[:hash_size]
    return f"{name[: (max_length - hash_size - 1)]}_{name_hash}"


def get_server_task_definition_family(
    task_definition_prefix: str,
    organization_name: Optional[str],
    deployment_name: str,
    location_name: str,
) -> str:
    # Truncate the location name if it's too long (but add a unique suffix at the end so that no matter what it's unique)
    # Relies on the fact that org name and deployment name are always <= 64 characters long to
    # stay well underneath the 255 character limit imposed by ECS
    m = hashlib.sha1()
    m.update(location_name.encode("utf-8"))

    # '{16}_{64}_{64}_{64}': max 211 characters
    truncated_location_name = _get_family_hash(location_name, max_length=64)

    final_family: str = (
        f"{task_definition_prefix}_{organization_name}_{deployment_name}_{truncated_location_name}"
    )

    assert len(final_family) <= 255

    return sanitize_family(final_family)


def get_run_task_definition_family(
    task_definition_prefix: str,
    organization_name: Optional[str],
    deployment_name: str,
    job_origin: RemoteJobOrigin,
) -> str:
    # Truncate the location name if it's too long (but add a unique suffix at the end so that no matter what it's unique)
    # Relies on the fact that org name and deployment name are always <= 64 characters long to
    # stay well underneath the 255 character limit imposed by ECS
    job_name = job_origin.job_name
    repo_name = job_origin.repository_origin.repository_name
    location_name = job_origin.repository_origin.code_location_origin.location_name

    assert len(task_definition_prefix) <= 16
    assert len(str(organization_name)) <= 64
    assert len(deployment_name) <= 64

    # '{16}_{64}_{64}_{32}_{32}_{32}': max 245 characters

    final_family = f"{task_definition_prefix}_{organization_name}_{deployment_name}_{_get_family_hash(location_name)}_{_get_family_hash(repo_name)}_{_get_family_hash(job_name)}"

    assert len(final_family) <= 255

    return sanitize_family(final_family)



import logging
import os
import subprocess
import sys
import threading
import time
from contextlib import AbstractContextManager
from typing import Optional, Union, cast

from dagster import _check as check
from dagster._core.errors import DagsterUserCodeUnreachableError
from dagster._core.instance.ref import InstanceRef
from dagster._grpc.client import DagsterGrpcClient, client_heartbeat_thread
from dagster._utils import find_free_port, safe_tempfile_path_unmanaged
from dagster._utils.error import SerializableErrorInfo, serializable_error_info_from_exc_info
from dagster_cloud_cli.core.workspace import CodeLocationDeployData, PexMetadata
from dagster_shared import seven
from dagster_shared.ipc import open_ipc_subprocess
from pydantic import BaseModel

from dagster_cloud.pex.grpc.server.registry import PexS3Registry
from dagster_cloud.pex.grpc.types import PexServerHandle
from dagster_cloud.workspace.user_code_launcher.utils import get_grpc_server_env

logger = logging.getLogger("dagster.multipex")


class PexProcessEntry(BaseModel, frozen=True, extra="forbid", arbitrary_types_allowed=True):
    pex_server_handle: PexServerHandle
    grpc_server_process: subprocess.Popen
    grpc_client: DagsterGrpcClient
    heartbeat_shutdown_event: threading.Event
    heartbeat_thread: threading.Thread


class PexErrorEntry(BaseModel, frozen=True, extra="forbid", arbitrary_types_allowed=True):
    pex_server_handle: PexServerHandle
    error: SerializableErrorInfo


class MultiPexManager(AbstractContextManager):
    def __init__(
        self,
        local_pex_files_dir: Optional[str] = None,
        watchdog_run_interval: Optional[int] = 30,
        enable_metrics: bool = False,
    ):
        # Keyed by hash of PexServerHandle
        self._pex_servers: dict[str, Union[PexProcessEntry, PexErrorEntry]] = {}
        self._pending_startup_pex_servers: set[str] = set()
        self._pending_shutdown_pex_servers: set[str] = set()
        self._pex_servers_lock = threading.RLock()
        self._pex_metadata_for_handle: dict[
            str, Optional[PexMetadata]
        ] = {}  # maps handle id to the pex tag
        self._heartbeat_ttl = 60
        self._registry = PexS3Registry(local_pex_files_dir)
        self._enable_metrics = enable_metrics

        if watchdog_run_interval and watchdog_run_interval > 0:
            self._watchdog_thread = threading.Thread(
                name="watchdog",
                target=self.run_watchdog_thread,
                args=(watchdog_run_interval,),
                daemon=True,
            )
            self._watchdog_thread.start()
            logger.info(
                "Created a watchdog thread %s for MultiPexManager with watchdog_run_interval=%s",
                self._watchdog_thread.name,
                watchdog_run_interval,
            )
        else:
            logger.info(
                "No watchdog thread started for MultiPexManager (watchdog_run_interval=%s)",
                watchdog_run_interval,
            )

    def run_watchdog_thread(self, interval: int) -> None:
        # Regularly check for unexpected terminations of all active code servers
        # and update status to error in _pex_servers
        while True:
            time.sleep(interval)
            active_servers = self.get_active_pex_servers()
            dead_server_returncodes = []
            for server in active_servers:
                server_id = server.pex_server_handle.get_id()
                returncode = server.grpc_server_process.poll()
                if returncode is not None:
                    dead_server_returncodes.append((server, returncode))
                    logger.error(
                        "watchdog: pex subprocesss %s unexpectedly exited with returncode %s -"
                        " changing state to error",
                        server_id,
                        returncode,
                    )
            self._mark_servers_unexpected_termination(dead_server_returncodes)
            if dead_server_returncodes:
                logger.warning(
                    "watchdog: inspected %s active servers %s, of which %s were found unexpectedly"
                    " terminated",
                    len(active_servers),
                    [server.pex_server_handle.get_id() for server in active_servers],
                    len(dead_server_returncodes),
                )

    def _mark_servers_unexpected_termination(
        self, dead_server_returncodes: list[tuple[PexProcessEntry, int]]
    ) -> None:
        with self._pex_servers_lock:
            for server, returncode in dead_server_returncodes:
                handle_id = server.pex_server_handle.get_id()
                if not self.is_server_active(handle_id):
                    # something already marked this server non active, do not overwrite the sate
                    continue
                self._pex_servers[handle_id] = PexErrorEntry(
                    pex_server_handle=server.pex_server_handle,
                    error=SerializableErrorInfo(
                        message=f"code server subprocess unexpectedly exited with {returncode}:",
                        cls_name=None,
                        stack=[],
                    ),
                )

    def get_pex_grpc_client_or_error(
        self, server_handle: PexServerHandle
    ) -> Union[DagsterGrpcClient, SerializableErrorInfo]:
        handle_id = server_handle.get_id()
        with self._pex_servers_lock:
            if handle_id not in self._pex_servers:
                try:
                    if handle_id in self._pending_startup_pex_servers:
                        raise Exception("This server is still starting up")
                    else:
                        raise Exception("No server created with the given handle")
                except Exception:
                    return serializable_error_info_from_exc_info(sys.exc_info())

            pex_server_or_error = self._pex_servers[handle_id]
            if isinstance(pex_server_or_error, PexErrorEntry):
                return pex_server_or_error.error

            return cast("PexProcessEntry", self._pex_servers[handle_id]).grpc_client

    def get_active_pex_servers(self) -> list[PexProcessEntry]:
        with self._pex_servers_lock:
            return [
                server
                for server_id, server in self._pex_servers.items()
                if self.is_server_active(server_id) and isinstance(server, PexProcessEntry)
            ]

    def get_error_pex_servers(self) -> list[PexErrorEntry]:
        with self._pex_servers_lock:
            return [
                server
                for server_id, server in self._pex_servers.items()
                if isinstance(server, PexErrorEntry)
            ]

    def get_active_pex_server_handles(
        self, deployment_name, location_name: str
    ) -> list[PexServerHandle]:
        return [
            server.pex_server_handle
            for server in self.get_active_pex_servers()
            if server.pex_server_handle.deployment_name == deployment_name
            and server.pex_server_handle.location_name == location_name
        ]

    def get_error_pex_server_handles(
        self, deployment_name, location_name: str
    ) -> list[PexServerHandle]:
        return [
            server.pex_server_handle
            for server in self.get_error_pex_servers()
            if server.pex_server_handle.deployment_name == deployment_name
            and server.pex_server_handle.location_name == location_name
        ]

    def get_all_pex_grpc_clients_map(self) -> dict[str, DagsterGrpcClient]:
        with self._pex_servers_lock:
            return {
                server.pex_server_handle.get_id(): server.grpc_client
                for server in self._pex_servers.values()
                if isinstance(server, PexProcessEntry)
            }

    def is_server_active(self, server_handle_id: str) -> bool:
        """Server is present and not pending shutdown."""
        with self._pex_servers_lock:
            return (
                server_handle_id in self._pex_servers
                and server_handle_id not in self._pending_shutdown_pex_servers
                and isinstance(self._pex_servers[server_handle_id], PexProcessEntry)
            )

    def create_pex_server(
        self,
        server_handle: PexServerHandle,
        code_location_deploy_data: CodeLocationDeployData,
        instance_ref: Optional[InstanceRef],
    ):
        # we keep track of pex metadata in use to help cleanup unused resources over time
        self._pex_metadata_for_handle[server_handle.get_id()] = (
            code_location_deploy_data.pex_metadata
        )

        # install pex files and launch them - do it asynchronously to not block this call
        def _create_pex_server() -> None:
            try:
                pex_executable = self._registry.get_pex_executable(
                    check.not_none(code_location_deploy_data.pex_metadata)
                )
                logger.info(
                    "Installed pex executable %s at %s",
                    code_location_deploy_data.pex_metadata,
                    pex_executable.source_path,
                )

                metadata = code_location_deploy_data
                logger.info("Launching subprocess %s", pex_executable.source_path)
                subprocess_args = [
                    pex_executable.source_path,
                    "-m",
                    "dagster",
                    "api",
                    "grpc",
                    "--heartbeat",
                    "--heartbeat-timeout",
                    str(self._heartbeat_ttl),
                ]
                if self._enable_metrics:
                    subprocess_args.append("--enable-metrics")

                # Set working_directory to "." if it is not set, so that it does not
                # default to the PEX absolute path (which will then be expected to be the
                # same during runs)
                if not metadata.working_directory:
                    metadata = metadata._replace(working_directory=".")

                if seven.IS_WINDOWS:
                    port = find_free_port()
                    socket = None
                else:
                    port = None
                    socket = safe_tempfile_path_unmanaged()

                additional_env = get_grpc_server_env(
                    code_location_deploy_data=metadata,
                    port=port,
                    location_name=server_handle.location_name,
                    instance_ref=instance_ref,
                    socket=socket,
                )

                server_process = open_ipc_subprocess(
                    subprocess_args,
                    env={
                        **os.environ.copy(),
                        **pex_executable.environ,
                        **additional_env,
                    },
                    cwd=pex_executable.working_directory,
                )

                client = DagsterGrpcClient(
                    port=port,
                    socket=socket,
                    host="localhost",
                    use_ssl=False,
                )

                heartbeat_shutdown_event = threading.Event()
                heartbeat_thread = threading.Thread(
                    target=client_heartbeat_thread,
                    args=(client, heartbeat_shutdown_event),
                    daemon=True,
                )
                heartbeat_thread.start()
                logger.info(
                    "Created a heartbeat thread %s for %s",
                    heartbeat_thread.name,
                    server_handle.get_id(),
                )

            except Exception:
                with self._pex_servers_lock:
                    self._pex_servers[server_handle.get_id()] = PexErrorEntry(
                        pex_server_handle=server_handle,
                        error=serializable_error_info_from_exc_info(sys.exc_info()),
                    )
                    self._pending_startup_pex_servers.remove(server_handle.get_id())
                logger.exception(
                    "Creating new pex server for %s:%s failed",
                    server_handle.deployment_name,
                    server_handle.location_name,
                )
                return

            with self._pex_servers_lock:
                self._pex_servers[server_handle.get_id()] = PexProcessEntry(
                    pex_server_handle=server_handle,
                    grpc_server_process=server_process,
                    grpc_client=client,
                    heartbeat_shutdown_event=heartbeat_shutdown_event,
                    heartbeat_thread=heartbeat_thread,
                )
                self._pending_startup_pex_servers.remove(server_handle.get_id())

        with self._pex_servers_lock:
            handle_id = server_handle.get_id()
            if handle_id in self._pending_startup_pex_servers:
                logger.info(
                    "Ignoring request to create pex server for %s - an identical server is"
                    " already pending start",
                    handle_id,
                )
                return
            self._pending_startup_pex_servers.add(handle_id)
            if handle_id in self._pex_servers:
                # clear any previous error state since we're attempting to start server again
                logger.info(
                    "Clearing previous state for %s: %s", handle_id, self._pex_servers[handle_id]
                )
                del self._pex_servers[handle_id]

        logger.info(
            "Creating new pex server for %s:%s",
            server_handle.deployment_name,
            server_handle.location_name,
        )
        threading.Thread(target=_create_pex_server).start()

    def shutdown_pex_server(self, server_handle: PexServerHandle):
        handle_id = server_handle.get_id()
        with self._pex_servers_lock:
            if handle_id in self._pex_servers or handle_id in self._pending_startup_pex_servers:
                logger.info("Server %s marked for shutdown", handle_id)
                self._pending_shutdown_pex_servers.add(handle_id)

    def cleanup_pending_shutdown_pex_servers(self) -> None:
        with self._pex_servers_lock:
            # clean up for any processes that have exited
            to_remove = set()
            for handle_id in self._pending_shutdown_pex_servers:
                if handle_id not in self._pex_servers:
                    continue
                server = self._pex_servers[handle_id]
                if isinstance(server, PexProcessEntry):
                    if server.grpc_server_process.poll() is not None:
                        # Server process shut down
                        to_remove.add(handle_id)
                    else:
                        try:
                            server.grpc_client.ping("")
                        except DagsterUserCodeUnreachableError:
                            logger.warning(
                                "server process is still running but the server is unreachable - killing the process",
                                exc_info=True,
                            )
                            server.grpc_server_process.kill()
                            to_remove.add(handle_id)

            for handle_id in to_remove:
                logger.info("Server %s completely shutdown, cleaning up", handle_id)
                self._pending_shutdown_pex_servers.remove(handle_id)
                del self._pex_servers[handle_id]

            # request shutdown for processes that we have not tried to shutdown yet
            for handle_id in self._pending_shutdown_pex_servers:
                pex_server = self._pex_servers.get(handle_id)
                if not pex_server:
                    # still in _pending_startup_pex_servers
                    logger.info("Server %s not up yet, will request shutdown later", handle_id)
                    continue

                if isinstance(pex_server, PexErrorEntry):
                    logger.debug("Server %s was in an error state, no shutdown needed", handle_id)
                    continue

                if pex_server.heartbeat_shutdown_event.is_set():
                    # already requested shutdown
                    continue

                logger.info("Requesting shutdown for server %s", handle_id)
                pex_server.heartbeat_shutdown_event.set()
                pex_server.heartbeat_thread.join()
                try:
                    pex_server.grpc_client.shutdown_server()
                except DagsterUserCodeUnreachableError:
                    logger.warning(
                        "Server shutdown for %s over grpc failed, killing the process",
                        handle_id,
                        exc_info=True,
                    )
                    pex_server.grpc_server_process.kill()

            # Delete any registry files not in use anymore
            # - ensure that resources for servers starting up or shutting down are not removed
            # - important to do this while holding the lock to avoid race conditions
            running_server_ids = {
                proc.pex_server_handle.get_id()
                for proc in self._pex_servers.values()
                if isinstance(proc, PexProcessEntry)
            }
            in_use_handle_ids = self._pending_startup_pex_servers.union(
                self._pending_shutdown_pex_servers
            ).union(running_server_ids)
            in_use_pex_metadatas = [
                self._pex_metadata_for_handle.get(handle_id) for handle_id in in_use_handle_ids
            ]
            in_use_pex_metadatas = [
                pex_metadata for pex_metadata in in_use_pex_metadatas if pex_metadata is not None
            ]
            self._registry.cleanup_unused_files(in_use_pex_metadatas)

    def __exit__(self, exception_type, exception_value, traceback):
        for pex_server in self._pex_servers.values():
            if isinstance(pex_server, PexProcessEntry):
                pex_server.heartbeat_shutdown_event.set()
                pex_server.heartbeat_thread.join()

        for pex_server in self._pex_servers.values():
            if isinstance(pex_server, PexProcessEntry):
                try:
                    pex_server.grpc_client.shutdown_server()
                except DagsterUserCodeUnreachableError:
                    # Server already shutdown
                    pass
                if pex_server.grpc_server_process.poll() is None:
                    pex_server.grpc_server_process.communicate(timeout=30)



import asyncio
import os
from collections.abc import Collection, Mapping, Sequence
from pathlib import Path
from typing import Any, Optional, cast

import boto3
import grpc
from dagster import (
    Array,
    Enum,
    EnumValue,
    Field,
    IntSource,
    Noneable,
    ScalarUnion,
    StringSource,
    _check as check,
)
from dagster._serdes import ConfigurableClass, ConfigurableClassData
from dagster._utils.merger import merge_dicts
from dagster_aws.ecs.container_context import EcsContainerContext
from dagster_aws.secretsmanager import get_secrets_from_arns

from dagster_cloud.api.dagster_cloud_api import UserCodeDeploymentType
from dagster_cloud.execution.monitoring import CloudContainerResourceLimits
from dagster_cloud.pex.grpc.types import GetCrashedPexServersArgs
from dagster_cloud.workspace.config_schema import SHARED_ECS_CONFIG
from dagster_cloud.workspace.ecs.client import (
    DEFAULT_ECS_GRACE_PERIOD,
    DEFAULT_ECS_TIMEOUT,
    ECS_EXEC_LINUX_PARAMETERS,
    Client,
    get_debug_ecs_prompt,
)
from dagster_cloud.workspace.ecs.run_launcher import CloudEcsRunLauncher
from dagster_cloud.workspace.ecs.service import Service
from dagster_cloud.workspace.ecs.utils import (
    get_ecs_human_readable_label,
    get_server_task_definition_family,
    unique_ecs_resource_name,
)
from dagster_cloud.workspace.user_code_launcher import (
    DEFAULT_SERVER_PROCESS_STARTUP_TIMEOUT,
    SHARED_USER_CODE_LAUNCHER_CONFIG,
    DagsterCloudGrpcServer,
    DagsterCloudUserCodeLauncher,
    ServerEndpoint,
)
from dagster_cloud.workspace.user_code_launcher.user_code_launcher import (
    UserCodeLauncherEntry,
    async_serialize_exceptions,
)
from dagster_cloud.workspace.user_code_launcher.utils import (
    deterministic_label_for_location,
    get_code_server_port,
    get_grpc_server_env,
)

EcsServerHandleType = Service

CONTAINER_NAME = "dagster"


class EcsUserCodeLauncher(DagsterCloudUserCodeLauncher[EcsServerHandleType], ConfigurableClass):
    def __init__(
        self,
        cluster: str,
        subnets: list[str],
        execution_role_arn: str,
        log_group: str,
        service_discovery_namespace_id: str,
        task_role_arn: Optional[str] = None,
        security_group_ids: Optional[list[str]] = None,
        inst_data: Optional[ConfigurableClassData] = None,
        secrets=None,
        secrets_tag=None,
        env_vars=None,
        ecs_timeout=None,
        ecs_grace_period=None,
        launch_type: Optional[str] = None,
        server_resources: Optional[Mapping[str, Any]] = None,
        run_resources: Optional[Mapping[str, Any]] = None,
        runtime_platform: Optional[Mapping[str, Any]] = None,
        mount_points: Optional[Sequence[Mapping[str, Any]]] = None,
        volumes: Optional[Sequence[Mapping[str, Any]]] = None,
        server_sidecar_containers: Optional[Sequence[Mapping[str, Any]]] = None,
        run_sidecar_containers: Optional[Sequence[Mapping[str, Any]]] = None,
        server_ecs_tags: Optional[Sequence[Mapping[str, Optional[str]]]] = None,
        run_ecs_tags: Optional[Sequence[Mapping[str, Optional[str]]]] = None,
        server_health_check: Optional[Mapping[str, Any]] = None,
        enable_ecs_exec=False,
        server_task_definition_prefix: str = "server",
        run_task_definition_prefix: str = "run",
        assign_public_ip: Optional[bool] = None,
        **kwargs,
    ):
        self.ecs = boto3.client("ecs")
        self.logs = boto3.client("logs")
        self.service_discovery = boto3.client("servicediscovery")
        self.secrets_manager = boto3.client("secretsmanager")

        self.cluster = cluster
        self.subnets = subnets
        self.security_group_ids = security_group_ids
        self.service_discovery_namespace_id = service_discovery_namespace_id
        self.execution_role_arn = execution_role_arn
        self.task_role_arn = task_role_arn
        self.log_group = log_group
        self._inst_data = check.opt_inst_param(inst_data, "inst_data", ConfigurableClassData)
        self.secrets = check.opt_list_param(secrets, "secrets")
        self.env_vars = check.opt_list_param(env_vars, "env_vars")

        if all(isinstance(secret, str) for secret in self.secrets):
            self.secrets = [
                {"name": name, "valueFrom": value_from}
                for name, value_from in get_secrets_from_arns(
                    self.secrets_manager, self.secrets
                ).items()
            ]

        self.secrets_tag = secrets_tag

        self._ecs_timeout = check.opt_int_param(
            ecs_timeout,
            "ecs_timeout",
            DEFAULT_ECS_TIMEOUT,
        )

        self._ecs_grace_period = check.opt_int_param(
            ecs_grace_period,
            "ecs_grace_period",
            DEFAULT_ECS_GRACE_PERIOD,
        )

        self.launch_type = check.opt_str_param(launch_type, "launch_type", default="FARGATE")

        self.server_resources = check.opt_mapping_param(server_resources, "server_resources")
        self.run_resources = check.opt_mapping_param(run_resources, "run_resources")

        self.runtime_platform = check.opt_mapping_param(runtime_platform, "runtime_platform")

        self.mount_points = check.opt_sequence_param(mount_points, "mount_points")
        self.volumes = check.opt_sequence_param(volumes, "volumes")

        self.server_sidecar_containers = check.opt_sequence_param(
            server_sidecar_containers, "server_sidecar_containers"
        )
        self.run_sidecar_containers = check.opt_sequence_param(
            run_sidecar_containers, "run_sidecar_containers"
        )

        self.server_task_definition_prefix = check.str_param(
            server_task_definition_prefix, "server_task_definition_prefix"
        )
        check.invariant(
            len(self.server_task_definition_prefix) <= 16,
            "server_task_definition_prefix must be at most 16 characters",
        )
        self.run_task_definition_prefix = check.str_param(
            run_task_definition_prefix, "run_task_definition_prefix"
        )

        check.invariant(
            len(self.run_task_definition_prefix) <= 16,
            "run_task_definition_prefix must be at most 16 characters",
        )

        self.server_ecs_tags = check.opt_sequence_param(server_ecs_tags, "server_ecs_tags")
        self.run_ecs_tags = check.opt_sequence_param(run_ecs_tags, "run_ecs_tags")

        self.server_health_check = check.opt_mapping_param(
            server_health_check, "server_health_check"
        )

        self._enable_ecs_exec = enable_ecs_exec

        self.client = Client(
            cluster_name=self.cluster,
            subnet_ids=self.subnets,
            security_group_ids=security_group_ids,
            service_discovery_namespace_id=self.service_discovery_namespace_id,
            log_group=self.log_group,
            show_debug_cluster_info=self.show_debug_cluster_info,
            timeout=self._ecs_timeout,
            grace_period=self._ecs_grace_period,
            launch_type=self.launch_type,
            assign_public_ip=assign_public_ip,
        )
        super().__init__(**kwargs)

    @property
    def show_debug_cluster_info(self) -> bool:
        return True

    @property
    def requires_images(self):
        return True

    @classmethod
    def config_type(cls):
        return merge_dicts(
            {
                "cluster": Field(StringSource),
                "subnets": Field(Array(StringSource)),
                "security_group_ids": Field(Array(StringSource), is_required=False),
                "execution_role_arn": Field(StringSource),
                "task_role_arn": Field(StringSource, is_required=False),
                "log_group": Field(StringSource),
                "service_discovery_namespace_id": Field(StringSource),
                "secrets": Field(
                    Array(
                        ScalarUnion(
                            scalar_type=str,
                            non_scalar_schema={"name": StringSource, "valueFrom": StringSource},
                        )
                    ),
                    is_required=False,
                    description=(
                        "An array of AWS Secrets Manager secrets. These secrets will "
                        "be mounted as environment variabls in the container. See "
                        "https://docs.aws.amazon.com/AmazonECS/latest/APIReference/API_Secret.html."
                    ),
                ),
                "secrets_tag": Field(
                    Noneable(StringSource),
                    is_required=False,
                    description=(
                        "AWS Secrets Manager secrets with this tag will be mounted as "
                        "environment variables in the container."
                    ),
                ),
                "env_vars": Field(
                    [StringSource],
                    is_required=False,
                    description=(
                        "List of environment variable names to include in the ECS task. Each can be"
                        " of the form KEY=VALUE or just KEY (in which case the value will be pulled"
                        " from the current process)"
                    ),
                ),
                "server_process_startup_timeout": Field(
                    IntSource,
                    is_required=False,
                    default_value=DEFAULT_SERVER_PROCESS_STARTUP_TIMEOUT,
                    description=(
                        "Timeout when waiting for a code server to be ready after it is created."
                        " You might want to increase this if your ECS tasks are successfully"
                        " starting but your gRPC server is timing out."
                    ),
                ),
                "ecs_timeout": Field(
                    IntSource,
                    is_required=False,
                    default_value=DEFAULT_ECS_TIMEOUT,
                    description=(
                        "How long (in seconds) to poll against ECS API endpoints. You might want to"
                        " increase this if your ECS tasks are taking too long to start up."
                    ),
                ),
                "ecs_grace_period": Field(
                    IntSource,
                    is_required=False,
                    default_value=DEFAULT_ECS_GRACE_PERIOD,
                    description=(
                        "How long (in seconds) to continue polling if an ECS API endpoint fails "
                        "(because the ECS API is eventually consistent)"
                    ),
                ),
                "launch_type": Field(
                    Enum(
                        "EcsLaunchType",
                        [
                            EnumValue("FARGATE"),
                            EnumValue("EC2"),
                        ],
                    ),
                    is_required=False,
                    default_value="FARGATE",
                    description=(
                        "What type of ECS infrastructure to launch the run task in. "
                        "See https://docs.aws.amazon.com/AmazonECS/latest/developerguide/launch_types.html"
                    ),
                ),
                "code_server_metrics": Field(
                    {"enabled": Field(bool, is_required=False, default_value=False)},
                    is_required=False,
                ),
                "agent_metrics": Field(
                    {"enabled": Field(bool, is_required=False, default_value=False)},
                    is_required=False,
                ),
                "enable_ecs_exec": Field(
                    bool,
                    is_required=False,
                    default_value=False,
                ),
                "server_task_definition_prefix": Field(
                    str, is_required=False, default_value="server"
                ),
                "run_task_definition_prefix": Field(
                    str, is_required=False, default_value="dagsterrun"
                ),
                "assign_public_ip": Field(
                    Noneable(bool),
                    is_required=False,
                    default_value=None,
                    description=(
                        "When using the FARGATE launch type, the launcher will attempt to automatically determine if it is "
                        "necessary to assign a public IP to the ECS task. In complex network topologies, this automatic "
                        "determination may not be accurate. In this case, you can explicitly set this value to True or False."
                    ),
                ),
            },
            SHARED_ECS_CONFIG,
            SHARED_USER_CODE_LAUNCHER_CONFIG,
        )

    @classmethod
    def from_config_value(cls, inst_data: ConfigurableClassData, config_value: dict[str, Any]):  # pyright: ignore[reportIncompatibleMethodOverride], fix me!
        return EcsUserCodeLauncher(inst_data=inst_data, **config_value)

    @property
    def inst_data(self) -> Optional[ConfigurableClassData]:
        return self._inst_data

    @property
    def user_code_deployment_type(self) -> UserCodeDeploymentType:
        return UserCodeDeploymentType.ECS

    def _write_readiness_sentinel(self) -> None:
        # Write to a sentinel file to indicate that we've finished our initial
        # reconciliation - this is used to indicate that we're ready to
        # serve requests
        Path("/opt/finished_initial_reconciliation_sentinel.txt").touch(exist_ok=True)
        self._logger.info(
            "Wrote readiness sentinel: indicating that agent is ready to serve requests"
        )

    def _get_grpc_server_sidecars(
        self, container_context: EcsContainerContext
    ) -> Optional[Sequence[Mapping[str, Any]]]:
        return container_context.server_sidecar_containers

    def _get_service_cpu_override(self, container_context: EcsContainerContext) -> Optional[str]:
        return container_context.server_resources.get("cpu")

    def _get_service_memory_override(self, container_context: EcsContainerContext) -> Optional[str]:
        return container_context.server_resources.get("memory")

    def _get_service_ephemeral_storage_override(
        self, container_context: EcsContainerContext
    ) -> Optional[int]:
        return container_context.server_resources.get("ephemeral_storage")

    def _get_service_code_server_replicas_count_override(
        self, container_context: EcsContainerContext
    ) -> Optional[int]:
        return container_context.server_resources.get("replica_count")

    def _get_service_repository_credentials_override(
        self, container_context: EcsContainerContext
    ) -> Optional[str]:
        return container_context.repository_credentials

    def _get_enable_ecs_exec(self) -> bool:
        return self._enable_ecs_exec

    def _get_additional_grpc_server_env(self) -> dict[str, str]:
        return {}

    def _get_dagster_tags(self, deployment_name: str, location_name: str) -> dict[str, str]:
        return {
            "dagster/deployment_name": get_ecs_human_readable_label(deployment_name),
            "dagster/location_name": get_ecs_human_readable_label(
                location_name,
            ),
            "dagster/location_hash": deterministic_label_for_location(
                deployment_name, location_name
            ),
        }

    def get_code_server_resource_limits(
        self, deployment_name: str, location_name: str
    ) -> CloudContainerResourceLimits:
        self._logger.info(
            f"Getting resource limits for {deployment_name}:{location_name}. resources: {self.server_resources}"
        )
        metadata = self._actual_entries[(deployment_name, location_name)].code_location_deploy_data
        resources = metadata.container_context.get("ecs", {}).get("server_resources", {})
        return {
            "ecs": {
                "cpu_limit": resources.get("cpu"),
                "memory_limit": resources.get("memory"),
            }
        }

    def _start_new_server_spinup(
        self,
        deployment_name: str,
        location_name: str,
        desired_entry: UserCodeLauncherEntry,
    ) -> DagsterCloudGrpcServer:
        metadata = desired_entry.code_location_deploy_data

        if metadata.pex_metadata:
            command = metadata.get_multipex_server_command(
                get_code_server_port(),
                metrics_enabled=self._instance.user_code_launcher.code_server_metrics_enabled,
            )
            additional_env = metadata.get_multipex_server_env()
            tags = {
                "dagster/multipex_server": "1",
                "dagster/agent_id": self._instance.instance_uuid,
                "dagster/server_timestamp": str(desired_entry.update_timestamp),
            }
        else:
            command = metadata.get_grpc_server_command(
                metrics_enabled=self._instance.user_code_launcher.code_server_metrics_enabled
            )
            additional_env = get_grpc_server_env(
                metadata,
                get_code_server_port(),
                location_name,
                self._instance.ref_for_deployment(deployment_name),
            )
            tags = {
                "dagster/grpc_server": "1",
                "dagster/agent_id": self._instance.instance_uuid,
                "dagster/server_timestamp": str(desired_entry.update_timestamp),
            }

        container_context = EcsContainerContext(
            secrets=self.secrets,
            secrets_tags=[self.secrets_tag] if self.secrets_tag else [],
            env_vars=self.env_vars
            + [f"{k}={v}" for k, v in (metadata.cloud_context_env or {}).items()],
            server_resources=self.server_resources,
            run_resources=self.run_resources,
            task_role_arn=self.task_role_arn,
            execution_role_arn=self.execution_role_arn,
            runtime_platform=self.runtime_platform,
            mount_points=self.mount_points,
            volumes=self.volumes,
            server_sidecar_containers=self.server_sidecar_containers,
            run_sidecar_containers=self.run_sidecar_containers,
            server_ecs_tags=self.server_ecs_tags,
            run_ecs_tags=self.run_ecs_tags,
            server_health_check=self.server_health_check,
        ).merge(EcsContainerContext.create_from_config(metadata.container_context))

        # disallow multiple replicas for code locations acting as pex servers
        if metadata.pex_metadata:
            replica_count = (
                metadata.container_context.get("ecs", {})
                .get("server_resources", {})
                .get("replica_count", None)
            )
            if replica_count and replica_count > 1:
                reason = "multiple replica_count are not supported for pex servers"
                raise Exception(
                    f"Cannot use multiple replicas for {deployment_name}:{location_name}: {reason}."
                )

        environment = merge_dicts(
            container_context.get_environment_dict(),
            additional_env,
            self._get_additional_grpc_server_env(),
            {"DAGSTER_GRPC_MAX_WORKERS": os.environ["DAGSTER_GRPC_MAX_WORKERS"]}
            if os.environ.get("DAGSTER_GRPC_MAX_WORKERS")
            else {},
        )

        self._logger.info(f"Creating a new service for {deployment_name}:{location_name}...")

        family = get_server_task_definition_family(
            self.server_task_definition_prefix,
            self._instance.organization_name,
            deployment_name,
            location_name,
        )

        system_tags = {**self._get_dagster_tags(deployment_name, location_name), **tags}
        system_tag_keys = set(system_tags)

        invalid_user_keys = [
            tag["key"] for tag in container_context.server_ecs_tags if tag["key"] in system_tag_keys
        ]
        if invalid_user_keys:
            raise Exception(f"Cannot override system ECS tags: {', '.join(invalid_user_keys)}")
        image = self._resolve_image(metadata)
        if image != metadata.image:
            self._logger.info("Resolved image to %r", image)

        service = self.client.create_service(
            name=unique_ecs_resource_name(deployment_name, location_name),
            family=family,
            image=image,
            container_name=CONTAINER_NAME,
            command=command,
            execution_role_arn=container_context.execution_role_arn,
            env=environment,
            tags={
                **self._get_dagster_tags(deployment_name, location_name),
                **tags,
                **{tag["key"]: tag.get("value") for tag in container_context.server_ecs_tags},
            },
            task_role_arn=container_context.task_role_arn,
            secrets=container_context.get_secrets_dict(self.secrets_manager),
            sidecars=self._get_grpc_server_sidecars(container_context),
            logger=self._logger,
            cpu=self._get_service_cpu_override(container_context),
            memory=self._get_service_memory_override(container_context),
            ephemeral_storage=self._get_service_ephemeral_storage_override(container_context),
            replica_count=self._get_service_code_server_replicas_count_override(container_context),
            repository_credentials=self._get_service_repository_credentials_override(
                container_context
            ),
            allow_ecs_exec=self._get_enable_ecs_exec(),
            runtime_platform=container_context.runtime_platform,
            mount_points=container_context.mount_points,
            volumes=container_context.volumes,
            health_check=container_context.server_health_check,
        )
        self._logger.info(
            f"Created a new service at hostname {service.hostname} for {deployment_name}:{location_name}, waiting for server to be ready..."
        )

        endpoint = ServerEndpoint(
            host=service.hostname,
            port=get_code_server_port(),
            socket=None,
        )

        return DagsterCloudGrpcServer(service, endpoint, metadata)

    def _check_running_multipex_server(self, multipex_server: DagsterCloudGrpcServer):
        self._logger.info(
            f"Checking whether service {multipex_server.server_handle.name} is ready for existing"
            " multipex server..."
        )
        asyncio.run(
            self.client.check_service_has_running_tasks(
                multipex_server.server_handle.name,
                container_name=CONTAINER_NAME,
                logger=self._logger,
            )
        )
        super()._check_running_multipex_server(multipex_server)

    @async_serialize_exceptions
    async def _wait_for_new_multipex_server(
        self,
        _deployment_name: str,
        _location_name: str,
        server_handle: Service,
        multipex_endpoint: ServerEndpoint,
    ):
        self._logger.info(
            f"Waiting for service {server_handle.name} to be ready for multipex server..."
        )
        task_arn = await self.client.wait_for_new_service(
            server_handle, container_name=CONTAINER_NAME, logger=self._logger
        )
        await self._wait_for_server_process(
            multipex_endpoint.create_multipex_client(),
            timeout=self._server_process_startup_timeout,
            additional_check=lambda: self.client.assert_task_not_stopped(
                task_arn, CONTAINER_NAME, self._logger
            ),
        )

    def _get_update_failure_debug_info(
        self,
        task_arn,
    ):
        sections = []

        try:
            logs = self.client.get_task_logs(task_arn, container_name=CONTAINER_NAME, limit=25)
            task_logs = "Task logs:\n" + "\n".join(logs) if logs else "No logs in task."
            sections.append(task_logs)
        except:
            self._logger.exception("Error trying to get logs for failed task", task_arn=task_arn)  # pyright: ignore[reportCallIssue]

        if self.show_debug_cluster_info:
            sections.append(get_debug_ecs_prompt(self.cluster, task_arn))

        return "\n\n".join(sections)

    async def _wait_for_new_server_ready(  # pyright: ignore[reportIncompatibleMethodOverride], fix me!
        self,
        deployment_name: str,
        location_name: str,
        user_code_launcher_entry: UserCodeLauncherEntry,
        server_handle: Service,
        server_endpoint: ServerEndpoint,
    ) -> None:
        metadata = user_code_launcher_entry.code_location_deploy_data
        self._logger.info(
            f"Waiting for service {server_handle.name} to be ready for gRPC server..."
        )
        task_arn = await self.client.wait_for_new_service(
            server_handle, container_name=CONTAINER_NAME, logger=self._logger
        )

        multipex_client = None

        if metadata.pex_metadata:
            multipex_server = check.not_none(
                self._get_multipex_server(deployment_name, location_name, metadata)
            )
            multipex_endpoint = multipex_server.server_endpoint
            multipex_client = multipex_endpoint.create_multipex_client()

        def _assert_new_pex_server_did_not_crash():
            self.client.assert_task_not_stopped(task_arn, CONTAINER_NAME, self._logger)

            crashed_pex_servers = []

            try:
                crashed_pex_servers = (
                    check.not_none(multipex_client)
                    .get_crashed_pex_servers(
                        GetCrashedPexServersArgs(
                            deployment_name=deployment_name,
                            location_name=location_name,
                        )
                    )
                    .server_handles
                )
            except Exception as e:
                if (
                    isinstance(e.__cause__, grpc.RpcError)
                    and cast("grpc.RpcError", e.__cause__).code() == grpc.StatusCode.UNIMPLEMENTED
                ):
                    # New gRPC method not implemented on old multipex server versions
                    pass
                else:
                    self._logger.exception("Error attempting to fetch crashed PEX servers")

            if any(
                pex_server_handle.metadata_update_timestamp
                == int(user_code_launcher_entry.update_timestamp)
                for pex_server_handle in crashed_pex_servers
            ):
                failure_debug_info = self._get_update_failure_debug_info(task_arn)
                raise Exception(
                    "Code server failed to start up."
                    + (f"\n\n{failure_debug_info}" if failure_debug_info else "")
                )

        if metadata.pex_metadata:
            additional_check = lambda: _assert_new_pex_server_did_not_crash()
        else:
            additional_check = lambda: self.client.assert_task_not_stopped(
                task_arn, CONTAINER_NAME, self._logger
            )

        await self._wait_for_dagster_server_process(
            client=server_endpoint.create_client(),
            timeout=self._server_process_startup_timeout,
            additional_check=additional_check,
            get_timeout_debug_info=lambda: self._get_update_failure_debug_info(task_arn),
        )

    def _remove_server_handle(self, server_handle: EcsServerHandleType) -> None:
        self._logger.info(
            f"Deleting service {server_handle.name} at hostname {server_handle.hostname}..."
        )
        self.client.delete_service(server_handle)
        self._logger.info(
            f"Deleted service {server_handle.name} at hostname {server_handle.hostname}."
        )

    def _get_multipex_server_handles_for_location(
        self, deployment_name, location_name: str
    ) -> Collection[EcsServerHandleType]:
        tags = {
            "dagster/location_hash": deterministic_label_for_location(
                deployment_name, location_name
            ),
            "dagster/multipex_server": "1",
            "dagster/agent_id": self._instance.instance_uuid,
        }
        return self.client.list_services(tags)

    def _get_standalone_dagster_server_handles_for_location(
        self, deployment_name, location_name: str
    ) -> Collection[EcsServerHandleType]:
        tags = {
            "dagster/location_hash": deterministic_label_for_location(
                deployment_name, location_name
            ),
            "dagster/grpc_server": "1",
            "dagster/agent_id": self._instance.instance_uuid,
        }
        return self.client.list_services(tags)

    def _list_server_handles(self) -> list[EcsServerHandleType]:
        return [
            service
            for service in self.client.list_services()
            if "dagster/location_name" in service.tags.keys()
        ]

    def get_agent_id_for_server(self, handle: EcsServerHandleType) -> Optional[str]:
        # Need to get container for server handle, then get the agent tag from that.
        return handle.tags.get("dagster/agent_id")

    def get_server_create_timestamp(self, handle: EcsServerHandleType) -> Optional[float]:
        return handle.create_timestamp

    def _run_launcher_kwargs(self) -> dict[str, Any]:
        return dict(
            task_definition={
                "log_group": self.log_group,
                "execution_role_arn": self.execution_role_arn,
                "requires_compatibilities": [self.launch_type],
                **({"task_role_arn": self.task_role_arn} if self.task_role_arn else {}),
                **(
                    {"sidecar_containers": self.run_sidecar_containers}
                    if self.run_sidecar_containers
                    else {}
                ),
                **({"runtime_platform": self.runtime_platform} if self.runtime_platform else {}),
                **({"mount_points": self.mount_points} if self.mount_points else {}),
                **({"volumes": self.volumes} if self.volumes else {}),
                **(
                    {"linux_parameters": ECS_EXEC_LINUX_PARAMETERS}
                    if self._get_enable_ecs_exec()
                    else {}
                ),
            },
            secrets=self.secrets,
            secrets_tag=self.secrets_tag,
            env_vars=self.env_vars,
            use_current_ecs_task_config=False,
            run_task_kwargs={
                "cluster": self.cluster,
                "networkConfiguration": self.client.network_configuration,
                "launchType": self.launch_type,
                **({"enableExecuteCommand": True} if self._get_enable_ecs_exec() else {}),
            },
            run_ecs_tags=self.run_ecs_tags,
            container_name=CONTAINER_NAME,
            run_resources=self.run_resources,
            task_definition_prefix=self.run_task_definition_prefix,
        )

    def run_launcher(self) -> CloudEcsRunLauncher:  # pyright: ignore[reportIncompatibleMethodOverride], fix me!
        launcher = CloudEcsRunLauncher(**self._run_launcher_kwargs())
        launcher.register_instance(self._instance)

        return launcher



import os
import time

import dagster._check as check
from dagster._core.launcher import RunLauncher
from dagster._core.launcher.base import LaunchRunContext
from dagster._core.utils import parse_env_var
from dagster._grpc.types import ExecuteRunArgs
from dagster_shared.ipc import open_ipc_subprocess

from dagster_cloud.execution.utils import TaskStatus
from dagster_cloud.execution.utils.process import check_on_process, kill_process
from dagster_cloud.workspace.user_code_launcher.utils import get_instance_ref_for_user_code

PID_TAG = "process/pid"


class CloudProcessRunLauncher(RunLauncher):
    def __init__(self):
        self._run_ids = set()
        super().__init__()

    def launch_run(self, context: LaunchRunContext) -> None:
        run = context.dagster_run
        pipeline_code_origin = check.not_none(context.job_code_origin)

        run_args = ExecuteRunArgs(
            job_origin=pipeline_code_origin,
            run_id=run.run_id,
            instance_ref=get_instance_ref_for_user_code(self._instance.get_ref()),
        )
        args = run_args.get_command_args()

        kwargs = {}
        if (
            run.job_code_origin
            and run.job_code_origin.repository_origin.container_context
            and run.job_code_origin.repository_origin.container_context.get("env_vars")
        ):
            kwargs["env"] = {**os.environ}
            for kev in run.job_code_origin.repository_origin.container_context["env_vars"]:
                key, value = parse_env_var(kev)
                kwargs["env"][key] = value

        p = open_ipc_subprocess(args, **kwargs)
        pid = p.pid

        self._run_ids.add(run.run_id)

        self._instance.add_run_tags(run.run_id, {PID_TAG: str(pid)})

    def join(self, timeout=30):
        total_time = 0
        interval = 0.01

        while True:
            active_run_ids = [
                run_id
                for run_id in self._run_ids
                if (
                    self._instance.get_run_by_id(run_id)
                    and not self._instance.get_run_by_id(run_id).is_finished  # pyright: ignore[reportOptionalMemberAccess]
                )
            ]

            if len(active_run_ids) == 0:
                return

            if total_time >= timeout:
                raise Exception(f"Timed out waiting for these runs to finish: {active_run_ids!r}")

            total_time += interval
            time.sleep(interval)
            interval = interval * 2

    def _get_pid(self, run):
        if not run or run.is_finished:
            return None

        tags = run.tags

        if PID_TAG not in tags:
            return None

        return int(tags[PID_TAG])

    def can_terminate(self, run_id):
        run = self._instance.get_run_by_id(run_id)
        if not run:
            return False

        pid = self._get_pid(run)
        if not pid:
            return False

        return check_on_process(pid) == TaskStatus.RUNNING

    def terminate(self, run_id):
        run = self._instance.get_run_by_id(run_id)
        if not run or run.is_finished:
            return False

        self._instance.report_run_canceling(run)

        pid = self._get_pid(run)
        if not pid:
            return False

        kill_process(pid)

        return True



from collections.abc import Sequence

from dagster._utils.error import SerializableErrorInfo

DAGSTER_FRAMEWORK_SUBSTRINGS = [
    "/site-packages/dagster/",
    "/python_modules/dagster/dagster",
]


def remove_dagster_framework_lines_from_serializable_exc_info(error_info: SerializableErrorInfo):
    return error_info._replace(
        stack=remove_dagster_framework_lines_from_stack_trace(error_info.stack),
        cause=(
            remove_dagster_framework_lines_from_serializable_exc_info(error_info.cause)
            if error_info.cause
            else None
        ),
        context=(
            remove_dagster_framework_lines_from_serializable_exc_info(error_info.context)
            if error_info.context
            else None
        ),
    )


def remove_dagster_framework_lines_from_stack_trace(
    stack: Sequence[str],
) -> Sequence[str]:
    # Remove lines until you find the first non-dagster framework line

    for i in range(len(stack)):
        if not _line_contains_dagster_framework_file(stack[i]):
            return stack[i:]

    # Return the full stack trace if its all Dagster framework lines,
    # to not be left with an empty stack trace
    return stack


def _line_contains_dagster_framework_file(line: str):
    # stack trace line starts with something like
    # File "/usr/local/lib/python3.11/site-packages/dagster/_core/execution/plan/utils.py",
    split_by_comma = line.split(",")
    if not split_by_comma:
        return False

    file_portion = split_by_comma[0]
    return any(
        framework_substring in file_portion for framework_substring in DAGSTER_FRAMEWORK_SUBSTRINGS
    )



from collections.abc import Iterable, Iterator
from typing import TYPE_CHECKING, Any, Optional, Union

import yaml
from dagster import (
    AssetCheckEvaluation,
    AssetCheckResult,
    AssetExecutionContext,
    AssetKey,
    AssetMaterialization,
    AssetObservation,
    OpExecutionContext,
    Output,
)

from dagster_cloud.dagster_insights.insights_utils import (
    extract_asset_info_from_event,
    handle_raise_on_error,
)
from dagster_cloud.dagster_insights.snowflake.snowflake_utils import (
    OPAQUE_ID_SQL_SIGIL,
    build_opaque_id_metadata,
    marker_asset_key_for_job,
)

if TYPE_CHECKING:
    from dagster_dbt import DbtCliInvocation


@handle_raise_on_error("dbt_cli_invocation")
def dbt_with_snowflake_insights(
    context: Union[OpExecutionContext, AssetExecutionContext],
    dbt_cli_invocation: "DbtCliInvocation",
    dagster_events: Optional[
        Iterable[
            Union[
                Output,
                AssetMaterialization,
                AssetObservation,
                AssetCheckResult,
                AssetCheckEvaluation,
            ]
        ]
    ] = None,
    skip_config_check=False,
    record_observation_usage: bool = True,
) -> Iterator[
    Union[Output, AssetMaterialization, AssetObservation, AssetCheckResult, AssetCheckEvaluation]
]:
    """Wraps a dagster-dbt invocation to associate each Snowflake query with the produced
    asset materializations. This allows the cost of each query to be associated with the asset
    materialization that it produced.

    If called in the context of an op (rather than an asset), filters out any Output events
    which do not correspond with any output of the op.

    Args:
        context (AssetExecutionContext): The context of the asset that is being materialized.
        dbt_cli_invocation (DbtCliInvocation): The invocation of the dbt CLI to wrap.
        dagster_events (Optional[Iterable[Union[Output, AssetObservation, AssetCheckResult, AssetCheckEvaluation]]]):
            The events that were produced by the dbt CLI invocation. If not provided, it is assumed
            that the dbt CLI invocation has not yet been run, and it will be run and the events
            will be streamed.
        skip_config_check (bool): If true, skips the check that the dbt project config is set up
            correctly. Defaults to False.
        record_observation_usage (bool): If True, associates the usage associated with
            asset observations with that asset. Default is True.

    **Example:**

    .. code-block:: python

        @dbt_assets(manifest=DBT_MANIFEST_PATH)
        def jaffle_shop_dbt_assets(
            context: AssetExecutionContext,
            dbt: DbtCliResource,
        ):
            dbt_cli_invocation = dbt.cli(["build"], context=context)
            yield from dbt_with_snowflake_insights(context, dbt_cli_invocation)
    """
    if not skip_config_check:
        is_snowflake = dbt_cli_invocation.manifest["metadata"]["adapter_type"] == "snowflake"
        dbt_project_config = yaml.safe_load(
            (dbt_cli_invocation.project_dir / "dbt_project.yml").open("r")
        )
        # sanity check that the sigil is present somewhere in the query comment
        query_comment = dbt_project_config.get("query-comment")
        if query_comment is None:
            raise RuntimeError("query-comment is required in dbt_project.yml but it was missing")
        comment = query_comment.get("comment")
        if comment is None:
            raise RuntimeError(
                "query-comment.comment is required in dbt_project.yml but it was missing"
            )
        if OPAQUE_ID_SQL_SIGIL not in comment:
            raise RuntimeError(
                "query-comment.comment in dbt_project.yml must contain the string"
                f" '{OPAQUE_ID_SQL_SIGIL}'. Read the Dagster Insights docs for more info."
            )
        if is_snowflake:
            # snowflake throws away prepended comments
            if not query_comment.get("append"):
                raise RuntimeError(
                    "query-comment.append must be true in dbt_project.yml when using the Snowflake"
                    " adapter"
                )

    if dagster_events is None:
        dagster_events = dbt_cli_invocation.stream()

    asset_and_partition_key_to_unique_id: list[tuple[AssetKey, Optional[str], Any]] = []
    for dagster_event in dagster_events:
        if isinstance(
            dagster_event,
            (
                AssetMaterialization,
                AssetObservation,
                Output,
                AssetCheckResult,
                AssetCheckEvaluation,
            ),
        ):
            unique_id = dagster_event.metadata["unique_id"].value
            asset_key, partition = extract_asset_info_from_event(
                context, dagster_event, record_observation_usage
            )
            if not asset_key:
                asset_key = marker_asset_key_for_job(context.job_def)

            asset_and_partition_key_to_unique_id.append((asset_key, partition, unique_id))

        yield dagster_event

    run_results_json = dbt_cli_invocation.get_artifact("run_results.json")
    invocation_id = run_results_json["metadata"]["invocation_id"]

    for asset_key, partition, unique_id in asset_and_partition_key_to_unique_id:
        # must match the query-comment in dbt_project.yml
        opaque_id = f"{unique_id}:{invocation_id}"
        yield AssetObservation(
            asset_key=asset_key,
            partition=partition,
            metadata=build_opaque_id_metadata(opaque_id),
        )



from abc import ABC, abstractmethod
from collections.abc import Sequence
from typing import AbstractSet  # noqa: UP035

import dagster._check as check
from dagster._record import record
from dagster_shared.serdes import whitelist_for_serdes


@whitelist_for_serdes
@record
class JobSelection(ABC):
    code_location_name: str
    repository_name: str

    @abstractmethod
    def resolve_job_names(self) -> AbstractSet[str]:
        raise NotImplementedError()

    @staticmethod
    def names(
        names: Sequence[str], code_location_name: str, repository_name: str
    ) -> "NamesJobSelection":
        check.invariant(len(names) == 1, "Only one job name is supported at this time")
        return NamesJobSelection(
            job_names=names, code_location_name=code_location_name, repository_name=repository_name
        )


@whitelist_for_serdes
@record
class NamesJobSelection(JobSelection):
    job_names: Sequence[str]

    def resolve_job_names(self) -> AbstractSet[str]:
        return set(self.job_names)
